import mesa

from .model import LetterSpace
from .SimpleContinuousModule import SimpleCanvas


def letter_draw(agent):
    colortuple = set(agent.topicVec)
    color = "#" + "".join(format(int(round(val * 255)), "02x") for val in colortuple)
    probDict = {
        "Shape": "circle",
        "r": agent.numLettersReceived,
        "Filled": "true",
        "Color": color
    }
    return probDict


def node_draw(G):
    nodes = []
    node_label_to_node_id = {}
    node_id = 0
    # edge_id = 0
    for node_label, agents in G.nodes.data("agent"):
        node = G.nodes()[node_label]
        agent = agents[0]
        colortuple = set(agent.topicVec)
        nodes.append({
            "size": node["numLettersReceived"],
            "color": "#" + "".join(format(int(round(val * 255)), "02x") for val in colortuple),
            "tooltip": f"id: {agent.unique_id}<br>Topics: {agent.topicVec}",
        })
        node_label_to_node_id[node_label] = node_id
        node_id += 1

    portrayal = dict()
    portrayal["nodes"] = nodes
    edges = []
    for (source, target) in G.edges():
        sourceAgent = [agents for x, agents in G.nodes.data("agent") if x == source]
        agent = sourceAgent[0][0]
        edges.append({
            "source": node_label_to_node_id[source],
            "target": node_label_to_node_id[target],
            "color": '#D3D3D3',
            "width": G.nodes()[source]["numLettersSend"],
        }
        )
    portrayal["edges"] = edges

    return portrayal


letter_canvas = SimpleCanvas(letter_draw, 360, 720)

network_canvas = mesa.visualization.NetworkModule(node_draw, 360, 720)


model_params = {
    "population": mesa.visualization.Slider(
        "Number of persons",
        50,
        10,
        200,
        10,
        description="Choose how many persons to include in the model.",
    ),
    "updateTopic": mesa.visualization.Slider(
        "Strength of adoption",
        0.05,
        0.01,
        0.3,
        0.05,
        description="Choose how strongly letter sending changes ones topics.",
    ),
    "updateHelp": mesa.visualization.StaticText(
        "Higher value:<br/> A letter sender faster adopts to the topic of a receiver."
    ),
    "threshold": mesa.visualization.Slider(
        "Similarity threshold",
        0.5,
        0.0,
        1.0,
        0.1,
        description="Choose how similar two persons topics have to be, to send a letter.",
    ),
    "thresholdHelp": mesa.visualization.StaticText(
        "Higher value:<br/> Sending a letter is less likely."
    ),
    "width": 360,
    "height": 180,
    "moveRange": mesa.visualization.Slider(
        "Range for moving position",
        20,
        0,
        100,
        10,
        description="Choose the visibility range for finding potential locations to move to.",
    ),
    "letterRange": mesa.visualization.Slider(
        "Range for letter sending",
        50,
        0,
        150,
        10,
        description="Choose the visibility range for finding potential receipients.",
    ),
    "minSep": 3
}

server = mesa.visualization.ModularServer(
    LetterSpace, [letter_canvas, network_canvas], "Random Letters", model_params
)



import os
from tqdm import tqdm
import pandas as pd

from semanticlayertools.metric.prune import PruneNetwork

def prune(
    modelparameters: dict,
    network: tuple,
    columns: list,
    iterations=10,
    delAmounts=(0.1, 0.25, 0.5, 0.75, 0.9),
    delTypes=("degree", "unif")
):
    """Generate pruned networks from input.
    
    Assumes existence of columns "sender", "receiver", and "step".
    """
    runDf = pd.DataFrame(network, columns = columns)
    pruning = PruneNetwork(runDf)
    result = pruning.deleteFromNetwork(
        iterations=iterations,
        delAmounts=delAmounts,
        delTypes=delTypes
    )
    result = result.assign(**modelparameters)
    return result
    


import mesa
import numpy as np
import networkx as nx

from .agents import LetterAgent, LetterNode


class LetterSpace(mesa.Model):
    """
    Letter sending model class. Handles agent creation, placement and scheduling.

    Each agent has a personal topic vector representing
    the shares the agent has of topics 1 to 3 with values
    between 0 and 1.
    Each agent has two different ranges, moveRange for deciding
    to move to another agents position, and letterRange to find
    potential correspondence partners. In addition a threshold
    value determines the necessary similarity between topic vectors
    to send a letter or not.

    Received and send letters are kept track of with two additional
    internal variables.
    """

    def __init__(
        self,
        population=100,
        width=360,
        height=180,
        moveRange=20,
        letterRange=30,
        threshold=0.5,
        updateTopic=0.1,
        minSep=5
    ):
        """Create a new Letter model."""
        self.population = population
        self.letterLedger = list()
        self.schedule = mesa.time.RandomActivation(self)
        self.G = nx.DiGraph()
        self.G.add_nodes_from([x for x in range(self.population)])
        self.space = mesa.space.ContinuousSpace(width, height, True)
        self.grid = mesa.space.NetworkGrid(self.G)
        self.factors = dict(minSep=minSep, updateTopic=updateTopic, threshold=threshold, moveRange=moveRange, letterRange=letterRange)
        self.make_agents()
        self.running = True

    def make_agents(self):
        """
        Create self.population agents, with random positions and starting headings.
        """
        for i, node in enumerate(self.G.nodes()):
            x = self.random.random() * self.space.x_max
            y = self.random.random() * self.space.y_max
            pos = np.array((x, y))
            topicVec = np.array([np.random.random() for x in range(3)])  # Agent has random shares of topics 1 - 3 between 0 and 1.
            letter = LetterAgent(
                i,
                self,
                pos,
                topicVec,
                **self.factors
            )
            nx.set_node_attributes(self.G, {i: {'numLettersSend': 0, 'numLettersReceived': 0}})
            letterNode = LetterNode(
                i,
                self,
                topicVec,
            )
            self.space.place_agent(letter, pos)
            self.grid.place_agent(letterNode, node)
            self.schedule.add(letter)

    def step(self):
        self.schedule.step()
        nx.spring_layout(self.G)
    
    def run(self, n):
        """Run the model for n steps."""
        for _ in range(n):
            self.step()



import random
import pandas as pd
import geopandas as gpd
from shapely import contains, LineString


def createData(population: int, populationDistribution: str):
    """Create random coordinates of historically motivated choices.
    
    The routine samples a population sample based on estimated
    population density of that coordinate. 

    The original CSV dataset is retrieved from 
    https://doi.org/10.1371/journal.pone.0162678.s003
    """

    initial_population_choices = pd.read_csv(
        populationDistribution,
        encoding="latin1", index_col=0
    )

    # Calculate relative population ratio to estimated settlement area.
    # This will correspont to the probabilities to draw an agent from
    # these coordinates.
    relPop = []

    for idx ,row in initial_population_choices.iterrows():
        relPop.append(
            row["Area"]/row["Pop"]
        )

    initial_population_choices.insert(0, 'relPop', relPop)

    # Four costal cities can not be considered, since the modern NUTS regions
    # give zero overlap to their coordinates, leading to potential errors when 
    # agents move.
    excludeCoastal = ['Great Yarmouth', 'Kingston-upon-Hull', 'Calais', 'Toulon']
    initial_population_choices = initial_population_choices.query("~Settlement.isin(@excludeCoastal)")

    loc_probabilities = []
    loc_values = []
    for idx, row in initial_population_choices.iterrows():
        loc_probabilities.append(row["relPop"])
        loc_values.append(
            (row["longitude"], row["latitude"])
        )

    coordinates = random.choices(
        loc_values,
        loc_probabilities,
        k=population
    )

    data = pd.DataFrame(
        coordinates,
        columns=["longitude", "latitude"]
    )

    data.insert(
        0,
        'unique_id',
        [
            "P" + str(x) for x in list(range(population))
        ]
    )

    # Read the Geodataframe with EPSG:4326 projection.
    geodf = gpd.GeoDataFrame(
        data,
        geometry=gpd.points_from_xy(data.longitude, data.latitude),
        crs="EPSG:4326"
    )

    # Transform to EPSG:3857, since the NUTS shape files are in 
    # that projection.
    geodf = geodf.to_crs("EPSG:3857")

    return geodf


def getRegion(geometry, model):
    """Get region ID overlaping with input geometry.
    
    Might e.g. fail if line of connection crosses international
    waters, since there is no NUTS region assigned then.
    """
    regionID = [
                x.unique_id for x in model.regions if contains(x.geometry, geometry)
            ]
    if regionID:
        return regionID[0]
    else:
        raise IndexError(f"Can not find overlaping region to geometry {geometry}")


def getPositionOnLine(start, target, returnType="point"):
    """Interpolate movement along line between two given points.
    
    The amount of moving from start to target is random.
    """
    segment = LineString([start, target])
    newPos = segment.interpolate(random.random(), normalized=True)
    if returnType == "point":
        return newPos
    elif returnType == "coords":
        return [x for x in newPos.coords][0]


import random
from typing import Dict, DefaultDict, Set
from collections import defaultdict

import mesa
import mesa_geo as mg
from shapely.geometry import Point

from scicom.historicalletters.agents import RegionAgent, SenderAgent


class Nuts2Eu(mg.GeoSpace):
    """Define regions containing senders of letters.
    
    This is modified from a mesa-geo example, here
    https://github.com/projectmesa/mesa-examples/blob/main/gis/geo_schelling_points/geo_schelling_points/space.py
    """
    _id_region_map: Dict[str, RegionAgent]

    def __init__(self):
        super().__init__(warn_crs_conversion=True)
        self._id_region_map = {}

    def add_regions(self, agents):
        super().add_agents(agents)
        total_area = 0
        for agent in agents:
            self._id_region_map[agent.unique_id] = agent

    def add_sender_to_region(self, agent, region_id):
        agent.region_id = region_id
        self._id_region_map[region_id].add_sender(agent)

    def remove_sender_from_region(self, agent):
        self._id_region_map[agent.region_id].remove_sender(agent)
        agent.region_id = None

    def add_sender(self, agent: SenderAgent, regionID: str) -> None:
        super().add_agents([agent])
        self.add_sender_to_region(agent, regionID)
        
    def move_sender(
        self, agent: SenderAgent, pos: mesa.space.FloatCoordinate, regionID: str
    ) -> None:
        self.__remove_sender(agent)
        agent.geometry = pos
        self.add_sender(agent, regionID)

    def __remove_sender(self, agent: SenderAgent) -> None:
        super().remove_agent(agent)
        self.remove_sender_from_region(agent)




import mesa
import numpy as np


class LetterAgent(mesa.Agent):
    """An agent with fixed initial start position.
  
    Each agent has a personal topic vector representing
    the shares the agent has of topics 1 to 10 with values
    between 0 and 1.
    Each agent has two different ranges, moveRange for deciding
    to move to another agents position, and letterRange to find
    potential correspondence partners. In addition a threshold
    value determines the necessary similarity between topic vectors
    to send a letter or not.

    Received and send letters are kept track of with two additional
    internal variables.
    """

    def __init__(
        self,
        unique_id,
        model,
        pos,
        topicVec,
        updateTopic,
        moveRange,
        letterRange,
        minSep,
        threshold
    ):
        """Create letter with position, topic vector and parameters."""
        super().__init__(unique_id, model)
        self.pos = np.array(pos)
        self.topicVec = topicVec
        self.updateTopic = updateTopic
        self.threshold = threshold
        self.moveRange = moveRange
        self.letterRange = letterRange
        self.topicLedger = []
        self.lettersReceived = []
        self.numLettersReceived = 0
        self.lettersSend = []
        self.numLettersSend = 0

    def move(self, neighbors):
        """The agent can randomly move to neighboring positions."""
        if neighbors:
            weights = []
            possible_steps = []
            for n in neighbors:
                possible_steps.append(n.pos)
                weights.append(n.lettersReceived)
            move = np.random.choice([0, 0, 0, 0, 0, 0, 0, 1])
            if move == 1:
                new_position = self.random.choice(possible_steps)
                self.model.space.move_agent(self, new_position)
  
    def sendLetter(self, neighbors):
        """Sending a letter based on an urn model."""
        possibleRec = []
        senders = self.lettersReceived
        receivers = self.lettersSend
        possibleRec.extend(senders)
        possibleRec.extend(receivers)
        if neighbors:
            neighborRec = []
            for n in neighbors:
                if n.numLettersReceived > 0:
                    nMult = [n.unique_id] * n.numLettersReceived
                    neighborRec.extend(nMult)
                else:
                    neighborRec.append(n.unique_id)
            possibleRec.extend(neighborRec)
        if possibleRec:
            recipientNr = np.random.choice(possibleRec)
            recipient = self.model.schedule.agents[recipientNr]
            topicChoices = self.topicLedger.copy()
            topicChoices.extend(recipient.topicLedger.copy())
            if topicChoices:
                randix = np.random.choice(list(range(len(topicChoices))))
                initTopic = topicChoices[randix]
            else:
                initTopic = self.topicVec
            distance = np.linalg.norm(recipient.topicVec - initTopic)
            if distance < self.threshold:
                recipient.numLettersReceived += 1
                recipient.lettersReceived.append(self.unique_id)
                self.numLettersSend += 1
                self.lettersSend.append(recipient.unique_id)
                # Update model social network
                self.model.G.add_edge(self.unique_id, recipient.unique_id)
                self.model.G.nodes()[self.unique_id]['numLettersSend'] = self.numLettersSend
                self.model.G.nodes()[recipient.unique_id]['numLettersReceived'] = recipient.numLettersReceived
                # Update agents topic vector
                updateTopicVec = self.topicVec + self.updateTopic * np.random.uniform(0, 1) * (recipient.topicVec - self.topicVec)
                self.model.letterLedger.append(
                    (self.unique_id, recipient.unique_id, self.pos, recipient.pos, updateTopicVec, self.model.schedule.steps)
                )
                self.topicLedger.append(
                    self.topicVec
                )
                self.topicVec = updateTopicVec
               
    def step(self):
        neighborsMove = self.model.space.get_neighbors(self.pos, self.moveRange, False)
        neighborsSend = self.model.space.get_neighbors(self.pos, self.letterRange, False)
        self.sendLetter(neighborsSend)
        self.move(neighborsMove)


class LetterNode(mesa.Agent):
    """An agent representing the network node.

    Only necessary for visualization purposes.
    """

    def __init__(
        self,
        unique_id,
        model,
        topicVec
    ):
        """Create letter with position, topic vector and parameters."""
        super().__init__(unique_id, model)
        self.topicVec = topicVec
        self.lettersReceived = 0
        self.lettersSend = 0



import argparse

parser = argparse.ArgumentParser(
    prog='Run SciCom Simulations',
    description='This script starts the server interfaces for the different ABM.',
    epilog='During the running of this script, the interfaces should be reachable at http://127.0.0.1:8521'
)

parser.add_argument(
    'simulation',
    choices=[
        "randomletters", "historicalletters", "knowledgespread"
    ]
)


args = parser.parse_args()

if args.simulation == "randomletters":
    from randomletters.server import server
    server.launch()
elif args.simulation == "historicalletters":
    from historicalletters.server import server
    server.launch()
elif args.simulation == "knowledgespread":
    from knowledgespread.server import server
    server.launch()



import mesa
import solara
import altair as alt
import pandas as pd
import nx_altair as nxa
import networkx as nx
import matplotlib.colors as colors

from scicom.knowledgespread.SimpleContinuousModule import SimpleCanvas
from mesa.visualization.modules import ChartVisualization


from scicom.knowledgespread.agents import ScientistAgent
from scicom.knowledgespread.model import KnowledgeSpread
from scicom.knowledgespread.utils import ageFunction

model_params = {
    "num_scientists": mesa.visualization.Slider(
        "Initial number of scientists",
        100, 10, 200, 10,
    ),
    "num_timesteps": mesa.visualization.Slider(
        "How long is the number of agents growing?",
        20, 5, 100, 5,
    ),
    "oppositionPercent": mesa.visualization.Slider(
        "Percentage of opposing agents",
        0.05, 0, 0.5, 0.05,
    ),
    "epiInit": mesa.visualization.Choice(
        'Choose initial conditions for epistemic space',
        value=('complex'),
        choices=['complex', "central", "polarized"]
    ),
    "timeInit": mesa.visualization.Choice(
        'Choose initial conditions for population growth.',
        value=('saturate'),
        choices=['saturate', "linear", "exponential"]
    ),
    "epiRange": mesa.visualization.Slider(
        "Basic range of visibility in epistemic space",
        0.005, 0.001, 0.03, 0.001,
    )
}


def agent_draw_altair(agent):
    shapeDict = {True: "triangle-up", False: "circle"}
    colortuple = set(agent.topicledger[-1])
    color = colors.to_hex(colortuple)
    #color = "#" + "".join(format(int(round(val * 255)), "02x") for val in colortuple)
    if agent.age > 0:
        size = ageFunction(agent, agent.a, agent.b, agent.c, 10)
    else:
        size = 0.001
    probDict = {
        #"Shape": shapeDict[agent.opposition],
        "opposition": agent.opposition,
        #"r": ageFunction(agent),
        "size": size,
        #"Filled": "true",
        "Color": color,
        "color": color
    }
    return probDict


def chart_draw_altair_agents(model):
    data = model.datacollector.get_model_vars_dataframe().reset_index()
    chart = (
        alt.Chart(data)
        .mark_bar(color="orange")
        .encode(
            x=alt.X("index", title="Step"),
            y=alt.Y(
                "Active Agents",
                title="Active agents"
            ),
        )
    )
    return solara.FigureAltair(chart)


def chart_draw_altair_communities(model):
    data = model.datacollector.get_model_vars_dataframe().reset_index()
    data.insert(0, "compLen", data["Graph structure"].apply(lambda x: len(x)))
    chart = (
        alt.Chart(data)
        .mark_bar(color="orange")
        .encode(
            x=alt.X("index", title="Step"),
            y=alt.Y(
                "compLen",
                title="Communities in social network"
            ),
        )
    )
    return solara.FigureAltair(chart)


def epiSpace_draw_altair(model, agent_portrayal):
    isOpposed = (False, True)
    shape_range = ("circle", "triangle-up")
    all_agent_data = []
    if not model.schedule.agents:
        return solara.Markdown("## Finished run")
    else:
        for agent in model.schedule.agents:
            cur_agent = agent_draw_altair(agent)
            cur_agent["x"] = agent.pos[0]
            cur_agent["y"] = agent.pos[1]
            all_agent_data.append(cur_agent)
        df = pd.DataFrame(all_agent_data)
        colors = list(set(a["color"] for a in all_agent_data))
        chart_color = alt.Color("color").legend(None).scale(domain=colors, range=colors)
        chart = (
            alt.Chart(df)
            .mark_point(filled=True)
            .encode(
                x=alt.X("x", axis=None),  # no x-axis label
                y=alt.Y("y", axis=None),  # no y-axis label
                size=alt.Size("size", title="current activity"),  # relabel size for legend
                color=chart_color,
                shape=alt.Shape(  # use shape to indicate choice
                    "opposition", scale=alt.Scale(domain=isOpposed, range=shape_range)
                ),
            )
            .configure_view(strokeOpacity=0)  # hide grid/chart lines
        )
        return solara.FigureAltair(chart)


def socialNetwork_draw_altair(model):
    currentTime = model.schedule.time        
    H = model.socialNetwork
    Graph = nx.Graph(((u, v, e) for u, v, e in H.edges(data=True) if e['time'] <= currentTime))
    pos = nx.kamada_kawai_layout(Graph)
    between = nx.betweenness_centrality(Graph)
    nx.set_node_attributes(Graph, between, 'between')
    communities = list(nx.community.label_propagation_communities(Graph))
    for f in Graph.nodes():
        for i, c in enumerate(communities):
            if f in c:
                Graph.nodes()[f].update(
                    {
                        "community": str(i),
                        "name": f
                    }
                )
    chart = nxa.draw_networkx(
        G=Graph,
        pos=pos,
        node_color="between",
        edge_color='time',
        cmap='viridis'
    )
    chart.configure_view(strokeOpacity=0)  # hide grid/chart lines
    return solara.FigureAltair(chart)


def agent_draw(agent):
    colortuple = set(agent.topicledger[-1])
    color = "#" + "".join(format(int(round(val * 255)), "02x") for val in colortuple)
    probDict = {
        "Shape": "circle",
        "r": ageFunction(agent),
        "Filled": "true",
        "Color": color
    }
    return probDict


chart = ChartVisualization.ChartModule(
    [
        {"Label": "Active Agents", "Color": "#000000"},
        {"Label": "Graph components", "Color": "black"}
    ],
    data_collector_name='datacollector',
)


epistemic_canvas = SimpleCanvas(agent_draw, 720, 720)


server = mesa.visualization.ModularServer(
    KnowledgeSpread,
    [epistemic_canvas, chart],
    "Knowledge spread",
    model_params,
)




import random
from pathlib import Path
#from statistics import mean
from tqdm import tqdm
from numpy import mean

import mesa
import networkx as nx
import mesa_geo as mg
from shapely import contains

from scicom.utilities.statistics import prune
from scicom.historicalletters.utils import createData

from scicom.historicalletters.agents import SenderAgent, RegionAgent
from scicom.historicalletters.space import Nuts2Eu


def getPrunedLedger(model):
    """Model reporter for simulation of archiving.
    
    Returns statistics of ledger network of model run
    and various iterations of statistics of pruned networks. 
    """
    # TODO: Add all model params
    if model.runPruning is True:
        ledgerColumns = ['sender', 'receiver', 'sender_location', 'receiver_location', 'topic', 'step']
        modelparams = {
            "population": model.population,
            "moveRange": model.moveRange,
            "letterRange": model.letterRange,
            "useActivation": model.useActivation,
            "useSocialNetwork": model.useSocialNetwork,
        }
        result = prune(
            modelparameters=modelparams,
            network=model.letterLedger,
            columns=ledgerColumns
        )
    else:
        result = model.letterLedger
    return result

def getComponents(model):
    """Model reporter to get number of components.
    
    The MultiDiGraph is converted to undirected, 
    considering only edges that are reciprocal, ie.
    edges are established if sender and receiver have 
    exchanged at least a letter in each direction.
    """
    newG = model.G.to_undirected(reciprocal=True)
    comp = nx.number_connected_components(newG)
    return comp


class HistoricalLetters(mesa.Model):
    """A letter sending model with historical informed initital positions.
    
    Each agent has an initial topic vector, expressed as a RGB value. The 
    initial positions of the agents is based on a weighted random draw
    based on data from [1]
    
    Each step, agents generate two neighbourhoods for sending letters and 
    potential targets to move towards. The probability to send letters is 
    a self-reinforcing process. During each sending the internal topic of 
    the sender is updated as a random rotation towards the receivers topic.

    [1] J. Lobo et al, Population-Area Relationship for Medieval European Cities,
        PLoS ONE 11(10): e0162678.
    """
    def __init__(
        self,
        population: int = 100,
        moveRange: float = 0.05,
        letterRange: float = 0.2,
        similarityThreshold: float = 0.2,
        updateTopic: float = 0.1,
        useActivation=False,
        useSocialNetwork=False,
        longRangeNetworkFactor=0.3,
        shortRangeNetworkFactor=0.8,
        runPruning=False,
        regionData: str = Path(Path(__file__).parent.parent.resolve(), "data/NUTS_RG_60M_2021_3857_LEVL_2.geojson"),
        populationDistributionData: str = Path(Path(__file__).parent.parent.resolve(), "data/pone.0162678.s003.csv"),
        tempfolder: str = "./",
        debug=False
    ):
        super().__init__()

        # Control variables
        self.population = population
        self.moveRange = moveRange
        self.letterRange = letterRange
        self.useActivation = useActivation
        # Initialize social network
        self.useSocialNetwork = useSocialNetwork
        self.G = nx.MultiDiGraph()
        self.longRangeNetworkFactor = longRangeNetworkFactor
        self.shortRangeNetworkFactor = shortRangeNetworkFactor
        # Collected output variables
        self.letterLedger = []
        self.runPruning = runPruning
        self.movements = 0
        self.updatedTopic = 0
        # Internal variables
        self.schedule = mesa.time.RandomActivation(self)
        self.scaleSendInput = {}
        self.updatedTopicsDict = {}
        self.space = Nuts2Eu()
        self.personRegionMap = {}
        self.tempfolder = tempfolder
        self.debug = debug

        initSenderGeoDf = createData(
            population,
            populationDistribution=populationDistributionData
        )

        # Calculate mean of mean distances for each agent. 
        # This is used as a measure for the range of exchanges.
        distances = []
        for idx, row in initSenderGeoDf.iterrows():
            p1 = row['geometry']
            distances.append(
                initSenderGeoDf.geometry.apply(lambda x: p1.distance(x)).mean()
            )
        self.meandistance = mean(distances)

        self.factors = dict(
            updateTopic=updateTopic,
            similarityThreshold=similarityThreshold,
            moveRange=moveRange,
            letterRange=letterRange,
        )

        # Set up the grid with patches for every NUTS region
        ac = mg.AgentCreator(RegionAgent, model=self)
        self.regions = ac.from_file(
            regionData,
            unique_id="NUTS_ID"
        )
        self.space.add_regions(self.regions)

        # Set up agent creator for senders
        ac_senders = mg.AgentCreator(
            SenderAgent,
            model=self,
            agent_kwargs=self.factors
        )

        # Create agents based on random coordinates generated 
        # in the createData step above, see util.py file.
        senders = ac_senders.from_GeoDataFrame(
            initSenderGeoDf,
            unique_id="unique_id"
        )

        # Create random set of initial topic vectors.
        topics = [
            tuple(
                [random.random() for x in range(3)]
            ) for x in range(self.population)
        ]

        # Attach topic and activationWeight to each agent,
        # connect to social network graph.
        for idx, sender in enumerate(senders):
            self.G.add_node(
                sender.unique_id,
                numLettersSend=0,
                numLettersReceived=0    
            )
            sender.topicVec = topics[idx]
            # Add current topic to dict
            self.updatedTopicsDict.update(
                {sender.unique_id: topics[idx]}
            )
            if useActivation is True:
                sender.activationWeight = random.random()

        for agent in senders:
            regionID = [
                x.unique_id for x in self.regions if contains(x.geometry, agent.geometry)
            ]
            try:
                self.space.add_sender(agent, regionID[0])
            except IndexError:
                raise IndexError(f"Problem finding region for {agent.geometry}.")
            self.schedule.add(agent)

        # Add graph to network grid for potential visualization.
        # TODO: Not yet implemented. Maybe use Solara backend for this? 
        self.grid = mesa.space.NetworkGrid(self.G)

        # Create social network
        if useSocialNetwork is True:
            for agent in self.schedule.agents:
                if isinstance(agent, SenderAgent):
                    self._createSocialEdges(agent, self.G)

        # TODO: What comparitive values are useful for visualizations?
        self.datacollector = mesa.DataCollector(
            model_reporters={
                "Ledger": getPrunedLedger,
                "Letters": lambda x: len(self.letterLedger),
                "Movements": lambda x: self.movements,
                "Clusters": getComponents
            },
        )

    def _createSocialEdges(self, agent, graph):
        """Create social edges with the different wiring factors.

        Define a close range by using the moveRange parameter. Among
        these neighbors, create a connection with probability set by
        the shortRangeNetworkFactor. 

        For all other agents, that are not in this closeRange group,
        create a connection with the probability set by the longRangeNetworkFactor.
        """
        closerange = [x for x in self.space.get_neighbors_within_distance(
            agent,
            distance=self.moveRange * self.meandistance,
            center=False
        ) if isinstance(x, SenderAgent)]
        for neighbor in closerange:
            if neighbor.unique_id != agent.unique_id:
                connect = random.choices(
                    population=[True, False],
                    weights=[self.shortRangeNetworkFactor, 1 - self.shortRangeNetworkFactor],
                    k=1
                )
                if connect[0] is True:
                    graph.add_edge(agent.unique_id, neighbor.unique_id, step=0)
                    graph.add_edge(neighbor.unique_id, agent.unique_id, step=0)
        longrange = [x for x in self.schedule.agents if x not in closerange and isinstance(x, SenderAgent)]
        for neighbor in longrange:
            if neighbor.unique_id != agent.unique_id:
                connect = random.choices(
                    population=[True, False],
                    weights=[self.longRangeNetworkFactor, 1 - self.longRangeNetworkFactor],
                    k=1
                )
                if connect[0] is True:
                    graph.add_edge(agent.unique_id, neighbor.unique_id, step=0)
                    graph.add_edge(neighbor.unique_id, agent.unique_id, step=0)

    def step(self):
        self.scaleSendInput.update(
            **{x.unique_id: x.numLettersReceived for x in self.schedule.agents}
        )
        self.schedule.step()
        self.datacollector.collect(self)

    def run(self, n):
        """Run the model for n steps."""
        if self.debug is True:
            for _ in tqdm(range(n)):
                self.step()
        else:
            for _ in range(n):
                self.step()



import math
from unittest.mock import Mock, patch
from collections import Counter
import pandas as pd
import geopandas as gpd
from pathlib import Path
import pytest

from scicom.historicalletters.model import (
    HistoricalLetters,
    createData,
    SenderAgent,
    prune,
)


def test_initial_population_creation(tmp_path):
    file = createData(
        population=30,
        populationDistribution=Path(Path(__file__).parent.parent.resolve(), "src/scicom/data/pone.0162678.s003.csv")
    )
    assert isinstance(file, gpd.GeoDataFrame)


def test_model_initialization():
    # initialize model for 30 agents with defaults
    model = HistoricalLetters(30)
    # 30 agents should be on the scheduler
    assert len(model.schedule.agents) == 30
